/*
 *  evioUtil.cc
 *
 *   Author:  Elliott Wolin, JLab, 11-dec-2006
*/


#ifdef vxworks
#include <iostream.h>
#endif
#include "evioUtil.hxx"

using namespace std;
using namespace evio;



//--------------------------------------------------------------
//-------------------- local utilities -------------------------
//--------------------------------------------------------------


namespace evio {

  /**
   * Only used internally, always returns true.
   * @return true Always!
   */
  static bool isTrue(const evioDOMNodeP pNode) {
    return(true);
  }
  
  
//--------------------------------------------------------------

}  // namespace evio


//-----------------------------------------------------------------------
//------------------------ evioStreamParser -----------------------------
//-----------------------------------------------------------------------


/**
 * Stream parses event in buffer.
 * @param buf Buffer containing event
 * @param handler evioStreamParserHandler object containing callbacks to handle container and leaf nodes
 * @param userArg Passed to handler callbacks
 * @return void* Pointer to root bank of new tree
 */
void *evioStreamParser::parse(const uint32_t *buf, 
                              evioStreamParserHandler &handler, void *userArg) throw(evioException) {
  
  if(buf==NULL)throw(evioException(0,"?evioStreamParser::parse...null buffer",__FILE__,__FUNCTION__,__LINE__));
  return((void*)parseBank(buf,BANK,0,handler,userArg));
}


//--------------------------------------------------------------


/**
 * Used internally to parse banks.
 * @param buf Buffer containing bank
 * @param bankType Bank type
 * @param depth Current depth in depth-first parse of event
 * @param handler evioStreamParserHandler object containing callbacks to handle container and leaf nodes
 * @param userArg Passed to handler callbacks
 * @return void* Used internally
 */
void *evioStreamParser::parseBank(const uint32_t *buf, int bankType, int depth, 
                                 evioStreamParserHandler &handler, void *userArg) throw(evioException) {

  int length,contentType,dataOffset,p,bankLen;
  uint16_t tag;
  uint8_t num;
  const uint32_t *data;
  uint32_t mask;

  void *newUserArg = userArg;


  /* get type-dependent info */
  switch(bankType) {

  case 0xe:
  case 0x10:
    length  	= buf[0]+1;
    tag     	= buf[1]>>16;
    contentType	= (buf[1]>>8)&0xff;
    num     	= buf[1]&0xff;
    dataOffset  = 2;
    break;

  case 0xd:
  case 0x20:
    length  	= (buf[0]&0xffff)+1;
    tag     	= buf[0]>>24;
    contentType = (buf[0]>>16)&0xff;
    num     	= 0;
    dataOffset  = 1;
    break;
    
  case 0xc:
  case 0x40:
    length  	= (buf[0]&0xffff)+1;
    tag     	= buf[0]>>20;
    contentType	= (buf[0]>>16)&0xf;
    num     	= 0;
    dataOffset  = 1;
    break;

  default:
    ostringstream ss;
    ss << hex << ios::showbase << bankType;
    throw(evioException(0,"?evioStreamParser::parseBank...illegal bank type: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
  }


  /* 
   * if a leaf node, call leaf handler.
   * if container node, call node handler and then parse contained banks.
   */
  void *newLeaf = NULL;
  switch (contentType) {

  case 0x0:
  case 0x1:
  case 0x2:
  case 0xb:
    // four-byte types
    newLeaf=handler.leafNodeHandler(length-dataOffset,tag,contentType,num,depth,&buf[dataOffset],userArg);
    break;

  case 0x3:
  case 0x6:
  case 0x7:
    // one-byte types
    newLeaf=handler.leafNodeHandler((length-dataOffset)*4,tag,contentType,num,depth,(int8_t*)(&buf[dataOffset]),userArg);
    break;

  case 0x4:
  case 0x5:
    // two-byte types
    newLeaf=handler.leafNodeHandler((length-dataOffset)*2,tag,contentType,num,depth,(int16_t*)(&buf[dataOffset]),userArg);
    break;

  case 0x8:
  case 0x9:
  case 0xa:
    // eight-byte types
    newLeaf=handler.leafNodeHandler((length-dataOffset)/2,tag,contentType,num,depth,(int64_t*)(&buf[dataOffset]),userArg);
    break;

  case 0xe:
  case 0x10:
  case 0xd:
  case 0x20:
  case 0xc:
  case 0x40:
    // container types
    newUserArg=handler.containerNodeHandler(length,tag,contentType,num,depth,userArg);


    // parse contained banks
    p       = 0;
    bankLen = length-dataOffset;
    data    = &buf[dataOffset];
    mask    = ((contentType==0xe)||(contentType==0x10))?0xffffffff:0xffff;

    depth++;
    while(p<bankLen) {
      parseBank(&data[p],contentType,depth,handler,newUserArg);
      p+=(data[p]&mask)+1;
    }
    depth--;

    break;


  default:
    ostringstream ss;
    ss << hex << ios::showbase << contentType;
    throw(evioException(0,"?evioStreamParser::parseBank...illegal content type: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
    break;

  }


  // new user arg is pointer to new node
  if(newLeaf!=NULL) {
    return(newLeaf);
  } else {
    return(newUserArg);
  }

}


//-----------------------------------------------------------------------------
//---------------------------- evioDOMNode ------------------------------------
//-----------------------------------------------------------------------------


/** 
 * Container node constructor used internally.
 * @param par Parent node
 * @param tag Node tag
 * @param num Node num
 * @param contentType Container node content type
 */
evioDOMNode::evioDOMNode(evioDOMNodeP par, uint16_t tag, uint8_t num, int contentType) throw(evioException)
  : parent(par), parentTree(NULL), contentType(contentType), tag(tag), num(num)  {
}


//-----------------------------------------------------------------------------


/** 
 * Static factory method to create container node.
 * @param tag Node tag
 * @param num Node num
 * @param cType Container node content type
 * @return Pointer to new node
 */
evioDOMNodeP evioDOMNode::createEvioDOMNode(uint16_t tag, uint8_t num, ContainerType cType) throw(evioException) {
  return(new evioDOMContainerNode(NULL,tag,num,cType));
}


//-----------------------------------------------------------------------------


/** 
 * Static factory method to create container node holding evioSerializable object.
 * @param tag Node tag
 * @param num Node num
 * @param o evioSerializable object
 * @param cType Container node content type
 * @return Pointer to new node
 */
evioDOMNodeP evioDOMNode::createEvioDOMNode(uint16_t tag, uint8_t num, const evioSerializable &o, ContainerType cType) 
  throw(evioException) {
  evioDOMContainerNode *c = new evioDOMContainerNode(NULL,tag,num,cType);
  o.serialize(c);
  return(c);
}


//-----------------------------------------------------------------------------


/** 
 * Static factory method to create container node using C function to fill container.
 * @param tag Node tag
 * @param num Node num
 * @param f C function that fills container node
 * @param userArg User arg passed to C function
 * @param cType Container node content type
 * @return Pointer to new node
 */
evioDOMNodeP evioDOMNode::createEvioDOMNode(uint16_t tag, uint8_t num, void (*f)(evioDOMNodeP c, void *userArg), 
                                            void *userArg, ContainerType cType) throw(evioException) {
  evioDOMContainerNode *c = new evioDOMContainerNode(NULL,tag,num,cType);
  f(c,userArg);
  return(c);
}


//-----------------------------------------------------------------------------


evioDOMNode::~evioDOMNode(void) {
}


//-----------------------------------------------------------------------------


/** 
 * Cleanly removes node from tree or node hierarchy.
 * @return Pointer to now liberated node
 */
evioDOMNodeP evioDOMNode::cut(void) throw(evioException) {
  if(parent!=NULL) {
    evioDOMContainerNode *par = static_cast<evioDOMContainerNode*>(parent);
    par->childList.remove(this);
    parent=NULL;
  } else if (parentTree!=NULL) {
    parentTree->root=NULL;
    parentTree=NULL;
  }
  return(this);
}


//-----------------------------------------------------------------------------


/** 
 * Cleanly removes node from tree or node hierarchy and recursively deletes node and its contents.
 */
void evioDOMNode::cutAndDelete(void) throw(evioException) {
  cut();
  delete(this);
}


//-----------------------------------------------------------------------------


/** 
 * Changes node parent.
 * @param newParent New parent
 * @return Pointer to now node just moved
 */
evioDOMNodeP evioDOMNode::move(evioDOMNodeP newParent) throw(evioException) {

  cut();

  evioDOMContainerNode *par = dynamic_cast<evioDOMContainerNode*>(newParent);
  if(par==NULL)throw(evioException(0,"?evioDOMNode::move...parent node not a container",__FILE__,__FUNCTION__,__LINE__));
  
  par->childList.push_back(this);
  parent=newParent;
  return(this);
}


//-----------------------------------------------------------------------------


/** 
 * Adds node to container node.
 * @param node Node to be added
 */
void evioDOMNode::addNode(evioDOMNodeP node) throw(evioException) {
  if(node==NULL)return;
  if(!isContainer())throw(evioException(0,"?evioDOMNode::addNode...not a container",__FILE__,__FUNCTION__,__LINE__));
  node->move(this);
}


//-----------------------------------------------------------------------------


/** 
 * Appends string to leaf node.
 * @param s string to append
 */
void evioDOMNode::append(const string &s) throw(evioException) {
  evioDOMLeafNode<string> *l = dynamic_cast<evioDOMLeafNode<string>*>(this);
  if(l!=NULL) {
    l->data.push_back(s);
  } else {
    throw(evioException(0,"?evioDOMNode::append...not appropriate node",__FILE__,__FUNCTION__,__LINE__));
  }
}


//-----------------------------------------------------------------------------


/** 
 * Appends const char* to leaf node.
 * @param s char* to append
 */
void evioDOMNode::append(const char *s) throw(evioException) {
  evioDOMLeafNode<string> *l = dynamic_cast<evioDOMLeafNode<string>*>(this);
  if(l!=NULL) {
    l->data.push_back(s);
  } else {
    throw(evioException(0,"?evioDOMNode::append...not appropriate node",__FILE__,__FUNCTION__,__LINE__));
  }
}


//-----------------------------------------------------------------------------


/** 
 * Appends char* to leaf node.
 * @param s char* to append
 */
void evioDOMNode::append(char *s) throw(evioException) {
  append((const char*)s);
}


//-----------------------------------------------------------------------------


/** 
 * Appends array of const char* to leaf node.
 * @param sa array of char* to append
 */
void evioDOMNode::append(const char **ca, int len) throw(evioException) {
  evioDOMLeafNode<string> *l = dynamic_cast<evioDOMLeafNode<string>*>(this);
  if(l!=NULL) {
    for(unsigned int i=0; i<len; i++) {
      l->data.push_back(ca[i]);
    }
  } else {
    throw(evioException(0,"?evioDOMNode::append...not appropriate node",__FILE__,__FUNCTION__,__LINE__));
  }
}


//-----------------------------------------------------------------------------


/** 
 * Appends array of const char* to leaf node.
 * @param sa array of char* to append
 */
void evioDOMNode::append(char **ca, int len) throw(evioException) {
  append((const char**)ca,len);
}


//-----------------------------------------------------------------------------


/** 
 * Appends single string to leaf node
 * @param sRef String to be added
 * @return Reference to this
 */
evioDOMNode& evioDOMNode::operator<<(const string &s) throw(evioException) {
  append(s);
  return(*this);
}


//-----------------------------------------------------------------------------


/** 
 * Appends single const char* to leaf node
 * @param s char* to be added
 * @return Reference to this
 */
evioDOMNode& evioDOMNode::operator<<(const char *s) throw(evioException) {
  append(s);
  return(*this);
}


//-----------------------------------------------------------------------------


/** 
 * Appends single char* to leaf node
 * @param s char* to be added
 * @return Reference to this
 */
evioDOMNode& evioDOMNode::operator<<(char *s) throw(evioException) {
  append((const char*)s);
  return(*this);
}


//-----------------------------------------------------------------------------


/** 
 * Returns pointer to child list of container node.
 */
evioDOMNodeList *evioDOMNode::getChildList(void) throw(evioException) {
  if(!isContainer())return(NULL);
  evioDOMContainerNode *c = dynamic_cast<evioDOMContainerNode*>(this);
  return(&c->childList);
}


//-----------------------------------------------------------------------------


/** 
 * Adds node to container node.
 * @param node Node to be added
 * @return Reference to node to be added
 */
evioDOMNode& evioDOMNode::operator<<(evioDOMNodeP node) throw(evioException) {
  addNode(node);
  return(*this);
}


//-----------------------------------------------------------------------------


/** 
 * True if node tag equals value.
 * @param tg Value to compare to
 * @return true if tags agree
 */
bool evioDOMNode::operator==(uint16_t tg) const {
  return(this->tag==tg);
}


//-----------------------------------------------------------------------------


/** 
 * True if node tag does not equal value
 * @param tg Value to compare to
 * @return true if tags disagree
 */
bool evioDOMNode::operator!=(uint16_t tg) const {
  return(this->tag!=tg);
}


//-----------------------------------------------------------------------------


/** 
 * True if node tag and num same as in tagNum pair.
 * @param tnPair tagNum pair to compare to
 * @return true if tag and num agree
 */
bool evioDOMNode::operator==(tagNum tnPair) const {
  return(
         (this->tag==tnPair.first) &&
         (this->num==tnPair.second)
         );
}


//-----------------------------------------------------------------------------


/** 
 * True if node tag and num NOT the same as in tagNum pair.
 * @param tnPair tagNum pair to compare to
 * @return true if tag and num disagree
 */
bool evioDOMNode::operator!=(tagNum tnPair) const {
  return(
         (this->tag!=tnPair.first) ||
         (this->num!=tnPair.second)
         );
}


//-----------------------------------------------------------------------------


/** 
 * Returns parent evioDOMNode.
 * @return Pointer to parent evioDOMNode
 */
const evioDOMNodeP evioDOMNode::getParent(void) const {
  return(parent);
}


//-----------------------------------------------------------------------------


/** 
 * Returns parent evioDOMTree.
 * @return Pointer to parent evioDOMTree
 */
const evioDOMTreeP evioDOMNode::getParentTree(void) const {
  return(parentTree);
}


//-----------------------------------------------------------------------------


/** 
 * Returns content type.
 * @return Content type
 */
int evioDOMNode::getContentType(void) const {
  return(contentType);
}


//-----------------------------------------------------------------------------


/** 
 * True if node is a container node.
 * @return true if node is a container
 */
bool evioDOMNode::isContainer(void) const {
  return(::is_container(contentType)==1);
}


//-----------------------------------------------------------------------------


/** 
 * True if node is a leaf node.
 * @return true if node is a leaf
 */
bool evioDOMNode::isLeaf(void) const {
  return(::is_container(contentType)==0);
}


//--------------------------------------------------------------


/** 
 * Returns indent for toString method, used internally
 * @return String containing proper number of indent spaces
 */
string evioDOMNode::getIndent(int depth) {
  string s;
  for(int i=0; i<depth; i++) s+="   ";
  return(s);
}


//-----------------------------------------------------------------------------
//----------------------- evioDOMContainerNode --------------------------------
//-----------------------------------------------------------------------------


/** 
 * Container node constructor used internally.
 * @param par Parent node
 * @param tg Node tag
 * @param num Node num
 * @param cType Container node content type
 */
evioDOMContainerNode::evioDOMContainerNode(evioDOMNodeP par, uint16_t tg, uint8_t num, ContainerType cType) throw(evioException)
  : evioDOMNode(par,tg,num,cType) {
}


//-----------------------------------------------------------------------------


/** 
 * Destructor recursively deletes children.
 */
evioDOMContainerNode::~evioDOMContainerNode(void) {
  evioDOMNodeList::iterator iter;
  for(iter=childList.begin(); iter!=childList.end(); iter++) {
    delete(*iter);
  }
}


//-----------------------------------------------------------------------------


/**
 * Returns XML string listing container node contents.
 * @return XML string listing contents
 */
string evioDOMContainerNode::toString(void) const {
  ostringstream os;
  os << getHeader(0) << getFooter(0);
  return(os.str());
}


//-----------------------------------------------------------------------------

                                   
/**
 * Returns XML string containing header needed by toString
 * @param depth Current depth
 * @return XML string
 */
string evioDOMContainerNode::getHeader(int depth) const {
  ostringstream os;
  os << getIndent(depth)
     <<  "<" << get_typename(parent==NULL?BANK:parent->getContentType()) << " content=\"" << get_typename(contentType)
     << "\" data_type=\"" << hex << ios::showbase << getContentType()
     << dec << "\" tag=\""  << tag;
  if((parent==NULL)||((parent->getContentType()==0xe)||(parent->getContentType()==0x10)))
    os << dec << "\" num=\"" << (int)num;
  os << "\">" << endl;
  return(os.str());
}


//-----------------------------------------------------------------------------

                                   
/**
 * Returns XML string containing footer needed by toString
 * @param depth Current depth
 * @return XML string
 */
string evioDOMContainerNode::getFooter(int depth) const {
  ostringstream os;
  os << getIndent(depth) << "</" << get_typename(this->parent==NULL?BANK:this->parent->getContentType()) << ">" << endl;
  return(os.str());
}


//-----------------------------------------------------------------------------
//------------------------------- evioDOMTree ---------------------------------
//-----------------------------------------------------------------------------


/**
 * Constructor fills tree from contents of evioChannel object.
 * @param channel evioChannel object
 * @param name Name of tree
 */
evioDOMTree::evioDOMTree(const evioChannel &channel, const string &name) throw(evioException) : root(NULL), name(name) {
  const uint32_t *buf = channel.getBuffer();
  if(buf==NULL)throw(evioException(0,"?evioDOMTree constructor...channel delivered null buffer",__FILE__,__FUNCTION__,__LINE__));
  root=parse(buf);
  root->parentTree=this;
}


//-----------------------------------------------------------------------------


/**
 * Constructor fills tree from contents of evioChannel object.
 * @param channel Pointer to evioChannel object
 * @param name Name of tree
 */
evioDOMTree::evioDOMTree(const evioChannel *channel, const string &name) throw(evioException) : root(NULL), name(name) {
  if(channel==NULL)throw(evioException(0,"?evioDOMTree constructor...null channel",__FILE__,__FUNCTION__,__LINE__));
  const uint32_t *buf = channel->getBuffer();
  if(buf==NULL)throw(evioException(0,"?evioDOMTree constructor...channel delivered null buffer",__FILE__,__FUNCTION__,__LINE__));
  root=parse(buf);
  root->parentTree=this;
}


//-----------------------------------------------------------------------------

/**
 * Constructor fills tree from contents of buffer
 * @param buf Buffer containing event
 * @param name Name of tree
 */
evioDOMTree::evioDOMTree(const uint32_t *buf, const string &name) throw(evioException) : root(NULL), name(name) {
  if(buf==NULL)throw(evioException(0,"?evioDOMTree constructor...null buffer",__FILE__,__FUNCTION__,__LINE__));
  root=parse(buf);
  root->parentTree=this;
}


//-----------------------------------------------------------------------------


/**
 * Constructor creates tree using node as the root node.
 * @param node Pointer to node that becomes the root node
 * @param name Name of tree
 */
evioDOMTree::evioDOMTree(evioDOMNodeP node, const string &name) throw(evioException) : root(NULL), name(name) {
  if(node==NULL)throw(evioException(0,"?evioDOMTree constructor...null evioDOMNode",__FILE__,__FUNCTION__,__LINE__));
  root=node;
  root->parentTree=this;
}


//-----------------------------------------------------------------------------


/**
 * Constructor creates new container node as root node.
 * @param tag Root node tag
 * @param num Root node num
 * @param cType Root node content type
 * @param name Name of tree
 */
evioDOMTree::evioDOMTree(uint16_t tag, uint8_t num, ContainerType cType, const string &name) throw(evioException) : root(NULL), name(name) {
  root=evioDOMNode::createEvioDOMNode(tag,num,cType);
  root->parentTree=this;
}


//-----------------------------------------------------------------------------


/**
 * Destructor deletes root node and contents.
 */
evioDOMTree::~evioDOMTree(void) {
  root->cutAndDelete();
}


//-----------------------------------------------------------------------------


/**
 * Parses contents of buffer and creates node hierarchry.
 * @param buf Buffer containing event data
 * @return Pointer to highest node resulting from parsing of buffer
 */
evioDOMNodeP evioDOMTree::parse(const uint32_t *buf) throw(evioException) {
  evioStreamParser p;
  return((evioDOMNodeP)p.parse(buf,*this,NULL));
}


//-----------------------------------------------------------------------------


/**
 * Creates container node, used internally when parsing buffer.
 * @param length not used
 * @param tag New node tag
 * @param contentType New node content type
 * @param num New node num
 * @param depth Current depth
 * @param userArg Used internally
 * @return void* used internally
 */
void *evioDOMTree::containerNodeHandler(int length, uint16_t tag, int contentType, uint8_t num, int depth, 
                               void *userArg) {
    
    
  // get parent pointer
  evioDOMContainerNode *parent = (evioDOMContainerNode*)userArg;
    

  // create new node
  evioDOMNodeP newNode = evioDOMNode::createEvioDOMNode(tag,num,(ContainerType)contentType);


  // set parent pointers
  if(parent!=NULL) {
    parent->childList.push_back(newNode);
    newNode->parent=parent;
  }
  

  // return pointer to new node
  return((void*)newNode);
}
  
  
//-----------------------------------------------------------------------------


/**
 * Creates leaf node, used internally when parsing buffer.
 * @param length Length of data in buffer
 * @param tag New node tag
 * @param contentType New node content type
 * @param num New node num
 * @param depth Current depth
 * @param data Pointer to data in buffer
 * @param userArg Used internally
 * @return void* Pointer to new leaf node
 */
void *evioDOMTree::leafNodeHandler(int length, uint16_t tag, int contentType, uint8_t num, int depth, 
                              const void *data, void *userArg) {


  // create and fill new leaf
  evioDOMNodeP newLeaf;
  string s;
  switch (contentType) {

  case 0x0:
  case 0x1:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(uint32_t*)data,length);
    break;
      
  case 0x2:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(float*)data,length);
    break;
      
  case 0x3:
    {
      newLeaf = evioDOMNode::createEvioDOMNode<string>(tag,num);
      char *start = (char*)data;
      char *c = start;
      while((c[0]!=0x4)&&((c-start)<length)) {
        s=string(c);
        cout << "string:  " << s << "  (len " << s.size() << ")" << endl;
        newLeaf->append(s);
        c+=s.size()+1;
      }
      break;
    }

  case 0x4:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(int16_t*)data,length);
    break;

  case 0x5:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(uint16_t*)data,length);
    break;

  case 0x6:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(int8_t*)data,length);
    break;

  case 0x7:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(uint8_t*)data,length);
    break;

  case 0x8:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(double*)data,length);
    break;

  case 0x9:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(int64_t*)data,length);
    break;

  case 0xa:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(uint64_t*)data,length);
    break;

  case 0xb:
    newLeaf = evioDOMNode::createEvioDOMNode(tag,num,(int32_t*)data,length);
    break;

  default:
    {
      ostringstream ss;
      ss << hex << ios::showbase << contentType;
      throw(evioException(0,"?evioDOMTree::leafNodeHandler...illegal content type: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
      break;
    }
  }


  // add new leaf to parent
  evioDOMContainerNode *parent = (evioDOMContainerNode*)userArg;
  if(parent!=NULL) {
    parent->childList.push_back(newLeaf);
    newLeaf->parent=parent;
  }


  // return pointer to newly created leaf node
  return((void*)newLeaf);
}


//-----------------------------------------------------------------------------


/** 
 * Removes and deletes tree root node and all its contents.
 */
void evioDOMTree::clear(void) throw(evioException) {
  if(root!=NULL) {
    root->cutAndDelete();
    root=NULL;
  }
}


//-----------------------------------------------------------------------------


/** 
 * Makes node root if tree is empty, or adds node to root if a container.
 * @param node Node to add to tree
 */
void evioDOMTree::addBank(evioDOMNodeP node) throw(evioException) {

  node->cut();

  if(root==NULL) {
    root=node;
    root->parentTree=this;

  } else {
    evioDOMContainerNode *c = dynamic_cast<evioDOMContainerNode*>(root);
    if(c==NULL)throw(evioException(0,"?evioDOMTree::addBank...root is not container",__FILE__,__FUNCTION__,__LINE__));
    c->childList.push_back(node);
    node->parent=root;
  }
}


//-----------------------------------------------------------------------------


/** 
 * Makes node root if tree is empty, or adds node to root if a container.
 * @param node Node to add to root
 */
evioDOMTree& evioDOMTree::operator<<(evioDOMNodeP node) throw(evioException) {
  addBank(node);
  return(*this);
}


//-----------------------------------------------------------------------------


/** 
 * Gets serialized length of tree.
 * @return Size of serialized tree in words
 */
int evioDOMTree::getSerializedLength(void) const throw(evioException) {
  return(getSerializedLength(root));
}


//-----------------------------------------------------------------------------


/** 
 * Serializes tree to buffer.
 * @param buf Buffer that receives serialized tree
 * @param size Size of buffer
 * @return Size of serialized buffer in words
 */
int evioDOMTree::toEVIOBuffer(uint32_t *buf, int size) const throw(evioException) {
  return(toEVIOBuffer(buf,root,size));
}


//-----------------------------------------------------------------------------


/** 
 * Gets serialized length of node, used internally.
 * @param pNode Node to serialize
 * @return Size of serialized node
 */
int evioDOMTree::getSerializedLength(const evioDOMNodeP pNode) const throw(evioException) {

  int bankLen,bankType,dataOffset;

  if(pNode->parent==NULL) {
    bankType=BANK;
  } else {
    bankType=pNode->parent->contentType;
  }


  // add bank header word(s)
  switch (bankType) {

  case 0xe:
  case 0x10:
    dataOffset=2;
    break;
  case 0xc:
  case 0xd:
  case 0x20:
  case 0x40:
    dataOffset=1;
    break;
  default:
    ostringstream ss;
    ss << hex << ios::showbase << bankType;
    throw(evioException(0,"evioDOMTree::getSerializedLength...illegal bank type: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
    break;
  }


  // set starting length
  bankLen=dataOffset;


  // loop over contained nodes if container node
  const evioDOMContainerNode *c = dynamic_cast<const evioDOMContainerNode*>(pNode);
  if(c!=NULL) {
    evioDOMNodeList::const_iterator iter;
    for(iter=c->childList.begin(); iter!=c->childList.end(); iter++) {
      bankLen+=getSerializedLength(*iter);
    }


  // leaf node...count data including padding
  } else {

    int nword,ndata;
    switch (pNode->contentType) {

    case 0x0:
    case 0x1:
    case 0x2:
    case 0xb:
      {
        const evioDOMLeafNode<uint32_t> *leaf = static_cast<const evioDOMLeafNode<uint32_t>*>(pNode);
        ndata = leaf->data.size();
        nword = ndata;
      }
      break;
      
    case 0x3:
      {
        const evioDOMLeafNode<string> *leaf = static_cast<const evioDOMLeafNode<string>*>(pNode);
        int nstring = leaf->data.size();
        ndata=0;
        for(unsigned int i=0; i<nstring; i++) ndata+=leaf->data[i].size();
        ndata+=nstring;        // account for nulls
        nword = (ndata+3)/4;   // include padding
      }
      break;

    case 0x4:
    case 0x5:
      {
        const evioDOMLeafNode<uint16_t> *leaf = static_cast<const evioDOMLeafNode<uint16_t>*>(pNode);
        ndata = leaf->data.size();
        nword = (ndata+1)/2;
      }
      break;

    case 0x6:
    case 0x7:
      {
        const evioDOMLeafNode<uint8_t> *leaf = static_cast<const evioDOMLeafNode<uint8_t>*>(pNode);
        ndata = leaf->data.size();
        nword = (ndata+3)/4;
      }
      break;

    case 0x8:
    case 0x9:
    case 0xa:
      {
        const evioDOMLeafNode<uint64_t> *leaf = static_cast<const evioDOMLeafNode<uint64_t>*>(pNode);
        ndata = leaf->data.size();
        nword = ndata*2;
      }
      break;

    default:
      ostringstream ss;
      ss << pNode->contentType;
      throw(evioException(0,"?evioDOMTree::getSerializedLength...illegal leaf type: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
      break;
    }

    // increment data count
    bankLen+=nword;
  }


  // return length of this bank
  return(bankLen);
}


//-----------------------------------------------------------------------------


/** 
 * Serializes node into buffer, used internally.
 * @param buf Buffer that receives serialized tree
 * @param pNode Node to serialize
 * @param size Size of buffer
 * @return Size of serialized node
 */
int evioDOMTree::toEVIOBuffer(uint32_t *buf, const evioDOMNodeP pNode, int size) const throw(evioException) {

  int bankLen,bankType,dataOffset;

  if(size<=0)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...illegal buffer size",__FILE__,__FUNCTION__,__LINE__));


  if(pNode->parent==NULL) {
    bankType=BANK;
  } else {
    bankType=pNode->parent->contentType;
  }


  // add bank header word(s)
  switch (bankType) {

  case 0xe:
  case 0x10:
    buf[0]=0;
    buf[1] = (pNode->tag<<16) | (pNode->contentType<<8) | pNode->num;
    dataOffset=2;
    break;
  case 0xd:
  case 0x20:
    if(pNode->num!=0)cout << "?warning...num ignored in segment: " << pNode->num << endl;
    buf[0] = (pNode->tag<<24) | ( pNode->contentType<<16);
    dataOffset=1;
    break;
  case 0xc:
  case 0x40:
    if(pNode->tag>0xfff)cout << "?warning...tag truncated to 12 bits in tagsegment: " << pNode->tag << endl;
    if(pNode->num!=0)cout << "?warning...num ignored in tagsegment: " << pNode->num<< endl;
    buf[0] = (pNode->tag<<20) | ( pNode->contentType<<16);
    dataOffset=1;
    break;
  default:
    ostringstream ss;
    ss << hex << ios::showbase << bankType;
    throw(evioException(0,"evioDOMTree::toEVIOBuffer...illegal bank type in boilerplate: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
    break;
  }


  // set starting length
  bankLen=dataOffset;


  // loop over contained nodes if container node
  const evioDOMContainerNode *c = dynamic_cast<const evioDOMContainerNode*>(pNode);
  if(c!=NULL) {
    evioDOMNodeList::const_iterator iter;
    for(iter=c->childList.begin(); iter!=c->childList.end(); iter++) {
      bankLen+=toEVIOBuffer(&buf[bankLen],*iter,size-bankLen-1);
      if(bankLen>size)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...buffer too small",__FILE__,__FUNCTION__,__LINE__));
    }


  // leaf node...copy data, and don't forget to pad!
  } else {

    int nword,ndata,i;
    switch (pNode->contentType) {

    case 0x0:
    case 0x1:
    case 0x2:
    case 0xb:
      {
        const evioDOMLeafNode<uint32_t> *leaf = static_cast<const evioDOMLeafNode<uint32_t>*>(pNode);
        ndata = leaf->data.size();
        nword = ndata;
        if(bankLen+nword>size)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...buffer too small",__FILE__,__FUNCTION__,__LINE__));
        for(i=0; i<ndata; i++) buf[dataOffset+i]=(uint32_t)(leaf->data[i]);
      }
      break;
      
    case 0x3:
      {
        const evioDOMLeafNode<string> *leaf = static_cast<const evioDOMLeafNode<string>*>(pNode);
        int nstring =  leaf->data.size();
        int nbytes,nbytesTotal=0;
        uint8_t *c = (uint8_t*)&buf[dataOffset];
        for(unsigned int j=0; j<nstring; j++) {
          string s = leaf->data[j];
          nbytes = s.size();
          nbytesTotal += nbytes + 1;
          if(bankLen+((nbytesTotal+3)/4)>size)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...buffer too small",__FILE__,__FUNCTION__,__LINE__));
          for(i=0; i<nbytes; i++) c[i]=(s.c_str())[i]; 
          c[nbytes]='\0';
          c+=nbytes+1;
        }
        for(i=0; i<(4-nbytesTotal%4)%4; i++) c[i]=0x04;   // pad with EOT if needed
        nword = (nbytesTotal+3)/4;
      }
      break;

    case 0x4:
    case 0x5:
      {
        const evioDOMLeafNode<uint16_t> *leaf = static_cast<const evioDOMLeafNode<uint16_t>*>(pNode);
        ndata = leaf->data.size();
        nword = (ndata+1)/2;
        if(bankLen+nword>size)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...buffer too small",__FILE__,__FUNCTION__,__LINE__));
        uint16_t *s = (uint16_t *)&buf[dataOffset];
        for(i=0; i<ndata; i++) s[i]=static_cast<uint16_t>(leaf->data[i]);
        if((ndata%2)!=0)buf[ndata]=0;
      }
      break;

    case 0x6:
    case 0x7:
      {
        const evioDOMLeafNode<uint8_t> *leaf = static_cast<const evioDOMLeafNode<uint8_t>*>(pNode);
        ndata = leaf->data.size();
        nword = (ndata+3)/4;
        if(bankLen+nword>size)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...buffer too small",__FILE__,__FUNCTION__,__LINE__));
        uint8_t *c = (uint8_t*)&buf[dataOffset];
        for(i=0; i<ndata; i++) c[i]=static_cast<uint8_t>(leaf->data[i]);
        for(i=ndata; i<ndata+(4-ndata%4)%4; i++) c[i]='\0';
      }
      break;

    case 0x8:
    case 0x9:
    case 0xa:
      {
        const evioDOMLeafNode<uint64_t> *leaf = static_cast<const evioDOMLeafNode<uint64_t>*>(pNode);
        ndata = leaf->data.size();
        nword = ndata*2;
        if(bankLen+nword>size)throw(evioException(0,"?evioDOMTree::toEVOIBuffer...buffer too small",__FILE__,__FUNCTION__,__LINE__));
        uint64_t *ll = (uint64_t*)&buf[dataOffset];
        for(i=0; i<ndata; i++) ll[i]=static_cast<uint64_t>(leaf->data[i]);
      }
      break;

    default:
      ostringstream ss;
      ss << pNode->contentType;
      throw(evioException(0,"?evioDOMTree::toEVOIBuffer...illegal leaf type: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
      break;
    }

    // increment data count
    bankLen+=nword;
  }


  // finaly set node length in buffer
  switch (bankType) {

  case 0xe:
  case 0x10:
    buf[0]=bankLen-1;
    break;
  case 0xd:
  case 0x20:
  case 0xc:
  case 0x40:
    if((bankLen-1)>0xffff)throw(evioException(0,"?evioDOMTree::toEVIOBuffer...length too big for segment type",__FILE__,__FUNCTION__,__LINE__));
    buf[0]|=(bankLen-1);
    break;
  default: 
    ostringstream ss;
    ss << bankType;
    throw(evioException(0,"?evioDOMTree::toEVIOBuffer...illegal bank type setting length: " + ss.str(),__FILE__,__FUNCTION__,__LINE__));
    break;
  }


  // return length of this bank
  return(bankLen);
}


//-----------------------------------------------------------------------------


/**
 * Returns list of all nodes in tree.
 * @return Pointer to list of nodes in tree (actually auto_ptr<>)
 */
evioDOMNodeListP evioDOMTree::getNodeList(void) throw(evioException) {
  return(evioDOMNodeListP(addToNodeList(root,new evioDOMNodeList,isTrue)));
}


//-----------------------------------------------------------------------------


/**
 * Returns XML string listing tree contents.
 * @return XML string listing contents
 */
string evioDOMTree::toString(void) const {

  if(root==NULL)return("<!-- empty tree -->");

  ostringstream os;
  os << endl << endl << "<!-- Dump of tree: " << name << " -->" << endl << endl;
  toOstream(os,root,0);
  os << endl << endl;
  return(os.str());

}


//-----------------------------------------------------------------------------


/**
 * Used internally by toString method.
 * @param os ostream to append XML fragments
 * @param pNode Node to get XML representation
 * @param depth Current depth
 */
void evioDOMTree::toOstream(ostream &os, const evioDOMNodeP pNode, int depth) const throw(evioException) {

  
  if(pNode==NULL)return;


  // get node header
  os << pNode->getHeader(depth);


  // dump contained banks if node is a container
  const evioDOMContainerNode *c = dynamic_cast<const evioDOMContainerNode*>(pNode);
  if(c!=NULL) {
    evioDOMNodeList::const_iterator iter;
    for(iter=c->childList.begin(); iter!=c->childList.end(); iter++) {
      toOstream(os,*iter,depth+1);
    }
  }


  // get footer
  os << pNode->getFooter(depth);
}


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
